﻿using Framework.changedetection;
using Framework.compression.algorithms;
using Framework.optimization;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Framework
{
    class Script
    {

        List<Task> tasks;
        float[] data;
        DeutscheBahn.DataSet dataset;

        DeutscheBahn.Job job;
        public string CompAlgoName { get { return tasks[0].CompressionAlgorithm; } }
        public string CPAlgoName { get { return tasks[0].CPAlgorithm; } }
        public string OptAlgoName { get { return tasks[0].OptimizationAlgorithm; } }
        public string DatasetName { get { return job.dataset.name; } }
        public string JobName { get { return job.name; } }
        public Script(DeutscheBahn.Job job)
        {
            this.job = job;
            this.tasks = JsonConvert.DeserializeObject<List<Task>>(job.json);
            this.data = job.dataset.DataArray;
            //Store the dataset if it's an initial parameter lookup task
            this.dataset = job.dataset;
        }

        // sender: main GUI
        public void Start()
        {

            // set job status to InProgress //
            int jobid = job.Id;

            foreach (Task task in tasks)
            {
                // raw data import from csv//
                //CsvImporter importer = new CsvImporter(task.Dataset);
                //importer.ReadFile();
                //Program.rawData = importer.GetColumn(0);

                Program.rawData = data;

                //this flag is to indicate in further process if initial changepoints are going to be searched! 
                //by default it's false  and only set to true
                bool paramlookup = false;

                // pick compression and initialize // 
                switch (task.CompressionAlgorithm)
                {
                    case "APCA": Program.CompAlgo = new APCA(); break;
                    case "SF": Program.CompAlgo = new SF(); break;
                    case "CHEB": Program.CompAlgo = new CHEB(); break;
                    case "PCA": Program.CompAlgo = new PCA(); break;
                    case "Wavelet": Program.CompAlgo = new Wavelet(); break;
                    case "Find Initial":
                        Program.CompAlgo = null;
                        paramlookup = true;
                        break;
                    default:
                        throw new ArgumentException("The CompAlgo was not found : \"" + task.CompressionAlgorithm + "\"");
                }
                Dictionary<string, RangedValue> parasTemp = null;

                if (!paramlookup) //only set compression params if not set
                {
                    // transfer parameters from task object to compression algo object //
                    if (task.CompressionInitialParameters != null)
                    {
                        parasTemp = Program.CompAlgo.GetParameters();
                        foreach (KeyValuePair<string, RangedValue> parameter in parasTemp)
                        {
                            if (task.CompressionInitialParameters.ContainsKey(parameter.Key))
                                parameter.Value.Value = task.CompressionInitialParameters[parameter.Key];
                        }
                        Program.CompAlgo.SetParameters(parasTemp);
                    }

                }
                // pick CP detection and initialize // 
                switch (task.CPAlgorithm)
                {
                    case "Bayes": Program.CpAlgo = new BayesOnlineCD(); break;
                    case "Kernel": Program.CpAlgo = new OnlineKernelChangeDetection(); break;
                    case "ChangeFinder": Program.CpAlgo = new ChangeFinderCD(); break;
                    case "ADWIN": Program.CpAlgo = new ADWIN(); break;
                    case "Direct Density-Ratio": Program.CpAlgo = new Framework.changedetection.DirectDensity(); break;
                    case "Event Detection": Program.CpAlgo = new EventDetection(); break;
                    default:
                        throw new ArgumentException("The CpAlgo was not found : \"" + task.CPAlgorithm + "\"");
                }


                Program.CpAlgo.SetDataset(Program.rawData);

                // transfer parameters from task object to CP algo object //
                if (task.CPInitialParameters != null)
                {
                    parasTemp = Program.CpAlgo.GetParameters();
                    foreach (KeyValuePair<string, RangedValue> parameter in parasTemp)
                    {
                        if (task.CPInitialParameters.ContainsKey(parameter.Key))
                            parameter.Value.Value = task.CPInitialParameters[parameter.Key];
                    }
                    Program.CpAlgo.SetParameters(parasTemp);
                }

                // pick optimization and initialize //
                IOptimizationAlgorithm optAlgo = null;
                switch (task.OptimizationAlgorithm)
                {
                    case "NM": optAlgo = new NelderMeadSolver();
                        break;
                    case "Evo": optAlgo = new EvolutionaryOptimization();
                        break;
                    case "NSGA2": optAlgo = new NSGA2();
                        break;
                    default:
                        throw new ArgumentException("The optimization algorithm was not found : \"" + task.OptimizationAlgorithm + "\"");
                }

                // transfer parameters from task object to optimization algo object //
                if (task.OptimizationParameters != null)
                {
                    parasTemp = optAlgo.GetParameters();
                    foreach (KeyValuePair<string, RangedValue> parameter in parasTemp)
                    {
                        if (task.OptimizationParameters.ContainsKey(parameter.Key))
                            parameter.Value.Value = task.OptimizationParameters[parameter.Key];
                    }
                    optAlgo.SetParameters(parasTemp);
                }

                Optimizer optimizer;
                DeutscheBahn.ChangePointSet changepointset;
                // do initial ChangePoint detection with Params which are set above //
                //this is only necessary if you want to compare to this initial set. In case of an param lookup we won't do this 
                if (!paramlookup)
                {
                    Program.CpAlgo.DetectChanges();
                    Program.rawChangePointScores = Program.CpAlgo.GetScores();
                    Program.rawChangePoints = Program.CpAlgo.GetChangePoints();
                    //These are the initial Changepoints for later comparison, Create a DB Conform Object!
                    changepointset = new DeutscheBahn.ChangePointSet();
                    changepointset.change_point = new List<DeutscheBahn.ChangePoint>();
                    foreach (int cp in Program.rawChangePoints)
                    {
                        DeutscheBahn.ChangePoint newcp = new DeutscheBahn.ChangePoint();
                        newcp.time = cp;
                        newcp.value = (Program.rawChangePointScores != null && Program.rawChangePointScores.Length > cp) ? Program.rawChangePointScores[cp] : 1;
                        changepointset.change_point.Add(newcp);
                    }

                    // initialize and optimizer object and set parameters //
                    optimizer = new Optimizer(Program.rawData, Program.CpAlgo, Program.CompAlgo, optAlgo);
                    optimizer.SetOptimizationParameters(task.OptFitnessParameters["alpha"],
                                task.OptFitnessParameters["beta"]);
                    optimizer.SetMeasurementParameters(task.OptFitnessParameters["a"],
                                    task.OptFitnessParameters["b"],
                                    task.OptFitnessParameters["c"]);
                }
                else
                {
                    //This optimization initialization finds the best init params
                    changepointset = new DeutscheBahn.ChangePointSet();
                    changepointset.change_point = new List<DeutscheBahn.ChangePoint>();
                    foreach (int cp in dataset.expectedCP) //Now create a result cpset from ground truth
                    {
                        DeutscheBahn.ChangePoint newcp = new DeutscheBahn.ChangePoint();
                        newcp.time = cp;
                        newcp.value = (dataset.expectedScores != null && dataset.expectedScores.Length > cp) ? dataset.expectedScores[cp] : 1;
                        changepointset.change_point.Add(newcp);
                    }
                    Program.CpAlgo.DetectChanges();
                    optimizer = new Optimizer(dataset.DataArray, dataset.expectedScores, dataset.expectedCP, Program.CpAlgo, optAlgo);
                    optimizer.SetOptimizationParameters(task.OptFitnessParameters["alpha"],
                                task.OptFitnessParameters["beta"]);
                    optimizer.SetMeasurementParameters(task.OptFitnessParameters["a"],
                                    task.OptFitnessParameters["b"],
                                    task.OptFitnessParameters["c"]);

                }


                // start optimization and wait for results. This step might take a while ... a short while ;) //
                DeutscheBahn.JobResult result = optimizer.createResult(optimizer.Solve());



                // write back results //
                using (var context = new DeutscheBahn.Model1Container())
                {
                    try
                    {
                        var query = (from jo in context.JobSet where jo.Id == jobid select jo).Take(1);
                        DeutscheBahn.Job meantJob = query.ToArray()[0];

                        meantJob.JobResult = result;
                        //now add all these crazy ChangePoints =) <3
                        meantJob.changepoint_set = changepointset;
                        meantJob.endtime = DateTime.Now;
                        meantJob.JobStatus = DeutscheBahn.Job.stat.Success;


                        context.SaveChanges();
                    }


                    catch (System.Data.Entity.Validation.DbEntityValidationException dbEx)
                    {

                        string datePatt = @"yyyy-MM-ddTHH.mm.ss";
                        String path = "RESULT_DUMP_" + jobid + "_" + DateTime.Now.ToString(datePatt) + ".txt";

                        StreamWriter sw = new StreamWriter(path);
                        sw.Write(JsonConvert.SerializeObject(result));
                        sw.Close();

                        Exception raise = dbEx;
                        foreach (var validationErrors in dbEx.EntityValidationErrors)
                        {
                            foreach (var validationError in validationErrors.ValidationErrors)
                            {
                                string message = string.Format("{0}:{1}",
                                    validationErrors.Entry.Entity.ToString(),
                                    validationError.ErrorMessage);
                                // raise a new exception nesting the current instance as InnerException
                                raise = new InvalidOperationException(message, raise);
                            }
                        }
                        throw raise;
                    }
                    catch (Exception)
                    {
                        string datePatt = @"yyyy-MM-ddTHH.mm.ss";
                        String path = "RESULT_DUMP_" + jobid + "_" + DateTime.Now.ToString(datePatt) + ".txt";
                        Console.WriteLine(path);
                        StreamWriter sw = new StreamWriter(path);
                        sw.Write(JsonConvert.SerializeObject(result));
                        sw.Close();
                    }
                }
            }
        }
    }

    class Task
    {
        public String Dataset;
        public String CompressionAlgorithm;
        public Dictionary<String, float> CompressionInitialParameters;
        public String CPAlgorithm;
        public Dictionary<String, float> CPInitialParameters;
        public String OptimizationAlgorithm;
        public Dictionary<String, float> OptimizationParameters;
        public Dictionary<String, float> OptFitnessParameters;
    }
}
