﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;

using MathWorks.MATLAB.NET.Arrays;
using MathWorks.MATLAB.NET.Utility;

namespace Framework.changedetection
{
    class BayesOnlineCD : IChangeDetectionAlgorithm
    {       
        private RangedFloatValue mu0;
        private RangedFloatValue kappa0;
        private RangedFloatValue alpha0;
        private RangedFloatValue beta0;
        private RangedFloatValue lambda;
        private float[] data;

        private float[] scores;
        private int[] changePoints;
        private bool hasScores;
        private bool hasRun = false;

        public BayesOnlineCD()
            : this(0, 1, 1, 1, 200)
        {
        }

        public BayesOnlineCD(float mu0, float kappa0, float alpha0, float beta0, float lambda)
        {
            this.mu0 = new RangedFloatValue(mu0, 0, 2);
            this.kappa0 = new RangedFloatValue(kappa0, 0, 5);
            this.alpha0 = new RangedFloatValue(alpha0, 0, 5);
            this.beta0 = new RangedFloatValue(beta0, 0, 5);
            this.lambda = new RangedFloatValue(lambda, Single.Epsilon, 500);
            this.hasScores = false;
        }

        public Dictionary<string, RangedValue> GetParameters()
        {
            Dictionary<string, RangedValue> parameters = new Dictionary<string, RangedValue>();
            parameters.Add("mu0", mu0);
            parameters.Add("kappa0", kappa0);
            parameters.Add("alpha0", alpha0);
            parameters.Add("beta0", beta0);
            parameters.Add("lambda", lambda);
            return parameters;
        }

        public void SetParameters(Dictionary<string, RangedValue> parameters)
        {
            if (parameters.ContainsKey("mu0"))
                this.mu0 = (RangedFloatValue)parameters["mu0"];
            if (parameters.ContainsKey("kappa0"))
                this.kappa0 = (RangedFloatValue)parameters["kappa0"];
            if (parameters.ContainsKey("alpha0"))
                this.alpha0 = (RangedFloatValue)parameters["alpha0"];
            if (parameters.ContainsKey("beta0"))
                this.beta0 = (RangedFloatValue)parameters["beta0"];
            if (parameters.ContainsKey("lambda"))
                this.lambda = (RangedFloatValue)parameters["lambda"];
        }

        public void SetDataset(float[] data)
        {
            this.data = data;
        }

        public void DetectChanges()
        {   
            MWArray[] result;
            try
            {
                BayesOnline.BayesOnlineClass detector = new BayesOnline.BayesOnlineClass();
                result = detector.BayesOnline(3,
                                              new MWNumericArray(this.data),
                                              new MWNumericArray((float)mu0.Value),
                                              new MWNumericArray((float)kappa0.Value),
                                              new MWNumericArray((float)alpha0.Value),
                                              new MWNumericArray((float)beta0.Value),
                                              new MWNumericArray((float)lambda.Value));
            }
            catch (TypeInitializationException e)
            {
                throw new TypeInitializationException("Could not initialize Matlab runtime.", e.InnerException);
            }

            changePoints = new int[result[2].NumberOfElements];
            int i = 0;
            foreach (double changePoint in result[2].ToArray())
            {
                changePoints[i] = (int)changePoint;
                i++;
            }

            scores = new float[result[1].NumberOfElements];
            i = 0;
            foreach (double score in result[1].ToArray())
            {
                scores[i] = (float)score;
                i++;
            }

            hasRun = true;
        }

        public bool HasScores()
        {
            return this.hasScores;
        }

        public float[] GetScores()
        {
            if (!hasRun)
            {
                throw new InvalidOperationException("Change detection was not executed. Call DetectChanges() first.");
            }

            float[] scores = new float[data.Length];
            for (int i = 0; i < changePoints.Length; i++)
            {
                scores[changePoints[i]] = 1;
            }
            return scores;
        }

        public int[] GetChangePoints()
        {
            if (!hasRun)
            {
                throw new InvalidOperationException("Change detection was not executed. Call DetectChanges() first.");
            }

            return this.changePoints;
        }


        public Dictionary<string, RangedValue> GetOptimizationParameters()
        {
            return GetParameters();
        }
        public string Name
        {
            get { return "Bayes"; }
        }
    }
}
