﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using MathWorks.MATLAB.NET.Arrays;
using MathWorks.MATLAB.NET.Utility;
using DirectDensityRatio = DirectDensity.DirectDensityRatio;

namespace Framework.changedetection
{
    class DirectDensity : IChangeDetectionAlgorithm
    {
        private RangedIntValue k; const string kK = "K";
        //private RangedIntValue n_rf; const string kN_rf = "n_rf";
        //private RangedIntValue n_te; const string kN_te = "n_te";
        private RangedIntValue n; const string kN = "n";
        private RangedFloatValue mu; const string kMu = "mu";
        private RangedFloatValue eta; const string kEta = "eta";
        private RangedFloatValue lambda; const string kLambda = "lambda";
        private float[] data;

        private float[] scores;
        private int[] changePoints;
        private bool hasScores;
        private bool hasRun = false;

        public DirectDensity()
        {
            this.k = new RangedIntValue(3, 3, 100);
            //this.n_rf = new RangedIntValue(10, 1, int.MaxValue);
            //this.n_te = new RangedIntValue(10, 1, int.MaxValue);
            this.n = new RangedIntValue(10, 3, 100);
            this.mu = new RangedFloatValue(1, 0, float.MaxValue);
            this.eta = new RangedFloatValue(0.1f, 0, 1f);
            this.lambda = new RangedFloatValue(0, 0, 1f);
            this.hasScores = true;
        }

        public Dictionary<string, RangedValue> GetParameters()
        {
            Dictionary<string, RangedValue> parameters = new Dictionary<string, RangedValue>();
            parameters.Add(kK, k);
            //parameters.Add(kN_rf, n_rf);
            //parameters.Add(kN_te, n_te);
            parameters.Add(kN, n);
            parameters.Add(kMu, mu);
            parameters.Add(kEta, eta);
            parameters.Add(kLambda, lambda);
            return parameters;
        }

        public void SetParameters(Dictionary<string, RangedValue> parameters)
        {
            if (parameters.ContainsKey(kK))
                this.k = (RangedIntValue)parameters[kK];
            //if (parameters.ContainsKey(kN_rf))
            //    this.n_rf = (RangedIntValue)parameters[kN_rf];
            //if (parameters.ContainsKey(kN_te))
            //    this.n_te = (RangedIntValue)parameters[kN_te];
            if (parameters.ContainsKey(kN))
                    this.n = (RangedIntValue)parameters[kN];
            if (parameters.ContainsKey(kMu))
                this.mu = (RangedFloatValue)parameters[kMu];
            if (parameters.ContainsKey(kEta))
                this.eta = (RangedFloatValue)parameters[kEta];
            if (parameters.ContainsKey(kLambda))
                this.lambda = (RangedFloatValue)parameters[kLambda];
        }

        public void SetDataset(float[] data)
        {
            this.data = data;
        }

        public void DetectChanges()
        {
            MWArray[] result;
            try
            {
                DirectDensityRatio detector = new DirectDensityRatio();
                result = detector.DirectDensity(2,
                                              new MWNumericArray(this.data),
                                              new MWNumericArray(k.Value),
                                              new MWNumericArray((float)n.Value),
                                              new MWNumericArray((float)n.Value),
                                              new MWNumericArray((float)mu.Value),
                                              new MWNumericArray((float)eta.Value),
                                              new MWNumericArray((float)lambda.Value));
            }
            catch (TypeInitializationException e)
            {
                throw new TypeInitializationException("Could not initialize Matlab runtime.", e.InnerException);
            }

            changePoints = new int[result[0].NumberOfElements];
            int i = 0;
            foreach (double changePoint in result[0].ToArray())
            {
                changePoints[i] = (int)changePoint;
                i++;
            }

            scores = new float[result[1].NumberOfElements];
            i = 0;
            foreach (double score in result[1].ToArray())
            {
                scores[i] = (float)score;
                i++;
            }

            hasRun = true;
        }

        public bool HasScores()
        {
            return this.hasScores;
        }

        public float[] GetScores()
        {
            if (!hasRun)
            {
                throw new InvalidOperationException("Change detection was not executed. Call DetectChanges() first.");
            }

            return this.scores;
        }

        public int[] GetChangePoints()
        {
            if (!hasRun)
            {
                throw new InvalidOperationException("Change detection was not executed. Call DetectChanges() first.");
            }

            return this.changePoints;
        }


        public Dictionary<string, RangedValue> GetOptimizationParameters()
        {
            return GetParameters();
        }

        public string Name
        {
            get { return "Direct Density-Ratio"; }
        }
    }
}
