﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;

using MathWorks.MATLAB.NET.Arrays;
using MathWorks.MATLAB.NET.Utility;

namespace Framework.changedetection
{
    class EventDetection : IChangeDetectionAlgorithm
    {
        private RangedFloatValue delta;
        private RangedIntValue windowSize;
        private RangedIntValue maxPolynom;
        private float[] data;
        private bool hasRun;

        private float[] scores;
        private int[] changePoints;


        public EventDetection()
        {
            this.delta = new RangedFloatValue(0.7f, 0 + Single.Epsilon, 1 - Single.Epsilon);
            this.windowSize = new RangedIntValue(50, 1, 200);
            this.maxPolynom = new RangedIntValue(3, 0, 5);
            this.hasRun = false;
        }


        public void SetParameters(Dictionary<string, RangedValue> parameters)
        {
            if (parameters.ContainsKey("delta"))
                this.delta = (RangedFloatValue)parameters["delta"];
            if (parameters.ContainsKey("windowSize"))
                this.windowSize = (RangedIntValue)parameters["windowSize"];
            if (parameters.ContainsKey("maxPolynom"))
                this.maxPolynom = (RangedIntValue)parameters["maxPolynom"];

            this.windowSize.Min = this.maxPolynom.Value + 1;
            if (this.windowSize.Value < this.windowSize.Min) {
                this.windowSize.Value = this.windowSize.Min;
            }
        }

        public Dictionary<string, RangedValue> GetParameters()
        {
            Dictionary<string, RangedValue> parameters = new Dictionary<string, RangedValue>();
            parameters.Add("delta", delta);
            parameters.Add("windowSize", windowSize);
            parameters.Add("maxPolynom", maxPolynom);
            return parameters;
        }

        public void SetDataset(float[] data)
        {
            this.data = data;
            this.windowSize.Max = (data.Length < 500) ? data.Length : 500;
        }

        public void DetectChanges()
        {
            MWArray[] result;

            try
            {
                eventDetection.EventDetectionClass detector = new eventDetection.EventDetectionClass();
                result = detector.eventDetection(2, new MWNumericArray((float)delta.Value), new MWNumericArray(this.maxPolynom.Value), new MWNumericArray(this.data), new MWNumericArray((int)windowSize.Value));

            }
            catch (TypeInitializationException e)
            {
                throw new TypeInitializationException("Could not initialize Matlab runtime.", e.InnerException);
            }

            changePoints = new int[result[0].NumberOfElements];
            int i = 0;
            foreach (double changePoint in result[0].ToArray())
            {
                changePoints[i] = (int)changePoint - 1;
                i++;
            }

            scores = new float[result[1].NumberOfElements];
            i = 0;
            foreach (double score in result[1].ToArray())
            {
                scores[i] = (float)score;
                i++;
            }

            hasRun = true;

        }

        public bool HasScores()
        {
            return true;
        }

        public float[] GetScores()
        {
            if (!hasRun)
            {
                throw new InvalidOperationException("Change detection was not executed. Call DetectChanges() first.");
            }

            return this.scores;
        }

        public int[] GetChangePoints()
        {
            if (!hasRun)
            {
                throw new InvalidOperationException("Change detection was not executed. Call DetectChanges() first.");
            }

            return this.changePoints;
        }


        public Dictionary<string, RangedValue> GetOptimizationParameters()
        {
            return GetParameters();
        }

        public string Name
        {
            get { return "Event Detection"; }
        }
    }
}
