﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using MathWorks.MATLAB.NET.Arrays;
using MathWorks.MATLAB.NET.Utility;
using System.Windows.Forms;
namespace Framework.changedetection
{
    class OnlineKernelChangeDetection : IChangeDetectionAlgorithm
    {
        Dictionary<string, RangedValue> algoParams = new Dictionary<String, RangedValue>();
        float[] dataset;

        private float[] scores;
        private int[] changePoints;
        private bool hasRun = false;

        public OnlineKernelChangeDetection()
        {
            algoParams.Add("m1", new RangedIntValue(50, 2, 200));
            algoParams.Add("m2", new RangedIntValue(50, 2, 200));
            algoParams.Add("nu", new RangedFloatValue(0.5f, 0.2f, 0.8f));
            algoParams.Add("eta", new RangedFloatValue(0.6f, 0, 1));
            algoParams.Add("sigma", new RangedFloatValue(0.1f, 0, 1));
        }

        public void SetParameters(Dictionary<string, RangedValue> parameters)
        {
            if (hasRun && parameters.Count == 1 && parameters.ContainsKey("eta"))
            {
                ((RangedFloatValue)algoParams["eta"]).Value = ((RangedFloatValue)parameters["eta"]).Value;
                calculateChangePoints();
            }
            foreach (KeyValuePair<string, RangedValue> kvp in parameters)
            {
                if (algoParams.ContainsKey(kvp.Key) && (kvp.Value.Value != algoParams[kvp.Key].Value))
                {
                    RangedValue oldValue = (RangedValue)algoParams[kvp.Key];
                    RangedValue newValue = (RangedValue)kvp.Value;
                    oldValue.Value = newValue.Value;
                    algoParams[kvp.Key] = oldValue;
                    hasRun = false;
                }
            }
        }

        public Dictionary<string, RangedValue> GetParameters()
        {
            return algoParams;
        }

        public void SetDataset(float[] data)
        {
            dataset = data;
            hasRun = false;
        }

        public void DetectChanges()
        {
            MWArray[] result;
            try
            {
                //Here goes the KernelChangeDetection fully Up
                OnlineKernel.KCD detector = new OnlineKernel.KCD();
                int m1 = (int)algoParams["m1"].Value;
                int m2 = (int)algoParams["m2"].Value;
                float nu = (float)algoParams["nu"].Value;
                float eta = (float)algoParams["eta"].Value;
                float sigma = (float)algoParams["sigma"].Value;
                result = detector.OnlineKernel(2, new MWNumericArray(this.dataset), m1, m2, nu, eta, sigma);
            }
            catch (TypeInitializationException e)
            {
                throw new TypeInitializationException("Could not initialize Matlab runtime.", e.InnerException);
            }

            changePoints = new int[result[1].NumberOfElements];
            int i = 0;
            foreach (double changePoint in result[1].ToArray())
            {
                changePoints[i] = (int)changePoint;
                i++;
            }

            scores = new float[result[0].NumberOfElements];
            i = 0;
            foreach (double score in result[0].ToArray())
            {
                scores[i] = (float)score;
                i++;
            }
            hasRun = true;
        }

        public bool HasScores()
        {
            return true;
        }

        public float[] GetScores()
        {
            if (!hasRun)
            {
                throw new InvalidOperationException("Change detection was not executed. Call DetectChanges() first.");
            }

            return this.scores;
        }

        public int[] GetChangePoints()
        {
            if (!hasRun)
            {
                throw new InvalidOperationException("Change detection was not executed. Call DetectChanges() first.");
            }

            return this.changePoints;
        }

        private void calculateChangePoints()
        {
            List<int> newChangePoints = new List<int>();
            for (int i = 0; i < scores.Length; i++)
            {
                if (scores[i] > (float)algoParams["eta"].Value)
                {
                    newChangePoints.Add(i);
                }
            }
            changePoints = newChangePoints.ToArray();
        }


        public Dictionary<string, RangedValue> GetOptimizationParameters()
        {
            return GetParameters();
        }

        public string Name
        {
            get { return "Kernel"; }
        }
    }
}
