﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Framework.compression
{
    interface ICompressionAlgorithm
    {
        string Name { get; }
        
        /// <summary>
        /// Sets the parameters for the algorithm. Since each algorithm has its own parameters a string-float dictionary is used.
        /// </summary>
        /// <param name="parameters">A dictionary of parameter names and values.</param>
        void SetParameters(Dictionary<string, RangedValue> parameters);


        /// <summary>
        /// Returns the parameter set. If SetParameters has not been called before, the default set with all possible parameters is returned.
        /// </summary>
        Dictionary<String, RangedValue> GetParameters();

        /// <summary>
        /// Returns the parameter set which are relevant for the optimization. 
        /// This is a proper subset of GetParameters
        /// </summary>
        Dictionary<String, RangedValue> GetOptimizationParameters();

        /// <summary>
        /// Sets the data to compress.
        /// </summary>
        /// <param name="data">An array of data values</param>
        void SetDataset(float[] data);

        /// <summary>
        /// Run the compression algorithm. Also decompresses the data again to provide uniform access to raw data and compressed data.
        /// </summary>
        void CompressDataset();

        /// <summary>
        /// Returns the reached compression ratio
        /// </summary>
        /// <returns>Returns the compression ratio.</returns>
        float getCompressionRatio();

        

        /// <summary>
        /// Returns the compressed dataset after compressDataset was executed. If called before compressDataset was executed it will throw an exception.
        /// </summary>
        /// <returns>Returns the compressed dataset.</returns>
        float[] GetCompressedDataset();

        /// <summary>
        /// Returns the decompressed dataset after compressDataset was executed. If called before compressDataset was executed it will throw an exception.
        /// </summary>
        /// <returns>Returns the decompressed dataset.</returns>
        float[] GetDecompressedDataset();
    }
}
