﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using MathWorks.MATLAB.NET.Arrays;
using MathWorks.MATLAB.NET.Utility;

namespace Framework.compression.algorithms
{
    class Wavelet : ICompressionAlgorithm
    {
        private float[] rawData;
        private RangedFloatValue threshold;
        /// <summary>
        /// This parameter is called KEEPAPP in MATLAB.
        /// If KEEPAPP = 1, approximation coefficients are not thresholded; otherwise, they can be thresholded.
        /// </summary>
        private RangedIntValue approximationCoefficient;
        private String waveletTyp = "db3";
        private int hardThreshold; // is in Matlab typ string
        private int maxLevel;
        private float[] decompData;
        private float compressionRatio;

        public Wavelet()
        {
            this.threshold = new RangedFloatValue(0.05f, 0, 0.3f);
            this.approximationCoefficient = new RangedIntValue(1, 0, 1); // Todo: kA ob die Range sinnvoll ist, habe ich mir ausgedacht
            hardThreshold = 1;
            maxLevel = 2;
        }

        /* wavelet types
        Daubechies:      'db1' or 'haar', 'db2', ... ,'db10', ... , 'db45'
        Coiflets:        'coif1', ... , 'coif5'
        Symlets:         'sym2', ... , 'sym8', ... ,'sym45'
        Discrete Meyer:  'dmey'
        Biorthogonal:
        'bior1.1', 'bior1.3', 'bior1.5'
        'bior2.2', 'bior2.4', 'bior2.6', 'bior2.8'
        'bior3.1', 'bior3.3', 'bior3.5', 'bior3.7'
        'bior3.9', 'bior4.4', 'bior5.5', 'bior6.8'
        Reverse Biorthogonal:
        'rbio1.1', 'rbio1.3', 'rbio1.5'
        'rbio2.2', 'rbio2.4', 'rbio2.6', 'rbio2.8'
        'rbio3.1', 'rbio3.3', 'rbio3.5', 'rbio3.7'
        'rbio3.9', 'rbio4.4', 'rbio5.5', 'rbio6.8'
         
         typThreshold: "s" or "h"
         */
        public void SetParameters(Dictionary<string, RangedValue> parameters)
        {
            if (parameters.ContainsKey("threshold"))
                this.threshold = (RangedFloatValue)parameters["threshold"];
            if (parameters.ContainsKey("approximationCoefficient"))
                this.approximationCoefficient = (RangedIntValue)parameters["approximationCoefficient"];
            if (parameters.ContainsKey("hardThreshold"))
                this.hardThreshold = (int)parameters["hardThreshold"].Value;
            if (parameters.ContainsKey("maxLevel"))
                this.maxLevel = (int)parameters["maxLevel"].Value;
        }

        public Dictionary<string, RangedValue> GetParameters()
        {
            Dictionary<string, RangedValue> parameters = new Dictionary<string, RangedValue>();
            parameters.Add("threshold", threshold);
            parameters.Add("approximationCoefficient", approximationCoefficient);
            parameters.Add("hardThreshold", new RangedIntValue(hardThreshold, 0, 1));
            parameters.Add("maxLevel", new RangedIntValue(maxLevel, 1, 10));
            return parameters;
        }

        public void SetDataset(float[] data)
        {
            rawData = data;
        }

        public void CompressDataset()
        {
            MLWavelet.WaveletClass wavelet = new MLWavelet.WaveletClass();
            MWNumericArray mwRawData = new MWNumericArray(rawData);
            MWCharArray mwTypThreshold = new MWCharArray((hardThreshold == 1) ? "h" : "s");
            MWCharArray mwWaveletTyp = new MWCharArray(waveletTyp);
            float epsilon = Utilities.calculateAbsoluteEpsilon(rawData, threshold.Value);
            MWArray[] mwDecompData = wavelet.waveletComp(2,
                                                         mwRawData,
                                                         mwWaveletTyp,
                                                         maxLevel,
                                                         new MWNumericArray(epsilon),
                                                         mwTypThreshold,
                                                         new MWNumericArray(approximationCoefficient.Value));

            decompData = new float[mwDecompData[0].NumberOfElements];
            for (int i = 0; i < mwDecompData[0].NumberOfElements; i++)
            {
                decompData[i] = Convert.ToSingle(mwDecompData[0].ToArray().GetValue(i, 0));
            }
            compressionRatio = Convert.ToSingle(mwDecompData[1].ToArray().GetValue(0, 0));
        }

        public float getCompressionRatio()
        {
            return compressionRatio;
        }

        public float[] GetCompressedDataset()
        {
            throw new NotImplementedException();
        }

        public float[] GetDecompressedDataset()
        {
            return decompData;
        }

        public Dictionary<string, RangedValue> GetOptimizationParameters()
        {
            return GetParameters();
        }

        public string Name
        {
            get { return "Wavelet"; }
        }
    }
}
