﻿using System;
using System.Collections.Generic;

using Framework.changedetection;
using Framework.compression;

namespace Framework.optimization
{
    /// <summary>
    /// Evolutionary optimization based on http://visualstudiomagazine.com/Articles/2014/02/01/Evolutionary-Optimization-Using-C.aspx
    /// </summary>
    class EvolutionaryOptimization : IOptimizationAlgorithm
    {
        const string kPopulationSize = "populationSize";
        const string kExitFitness = "exitFitness";
        const string kMaxGenerations = "maxGenerations";
        const string kMutateRate = "mutateRate";
        const string kMutateChange = "mutateChange";
        const string kTau = "tau";

        Random rnd = new Random();

        Individual[] population;

        /// <summary>
        /// Size of the population
        /// </summary>
        int populationSize = 50;
        /// <summary>
        /// Optimization is finishe when exitFitness is reached
        /// </summary>
        double exitFitness = 0.001;
        /// <summary>
        /// Optimization will abort after this many loops
        /// </summary>
        int maxGenerations = 500;
        /// <summary>
        /// Likelihood that a gene is changed
        /// </summary> 
        double mutateRate = 0.20;
        /// <summary>
        /// Controls magnitude of mutation change as percentage of total gene range
        /// </summary>
        double mutateChange = 0.01;
        /// <summary>
        /// Selection pressure (percentage of population selected for tournament selection)
        /// </summary>
        double tau = 0.40;

        private List<string> paramNames;

        public EvolutionaryOptimization()
        {
            population = new Individual[populationSize];
        }

        public void SetParameters(Dictionary<string, RangedValue> parameters)
        {
            foreach (KeyValuePair<string, RangedValue> kvp in parameters)
            {
                switch (kvp.Key)
                {
                    case kPopulationSize:
                        populationSize = (int)kvp.Value.Value;
                        population = new Individual[populationSize];
                        break;
                    case kExitFitness:
                        exitFitness = kvp.Value.Value;
                        break;
                    case kMaxGenerations:
                        maxGenerations = (int)kvp.Value.Value;
                        break;
                    case kMutateRate:
                        mutateRate = kvp.Value.Value;
                        break;
                    case kMutateChange:
                        mutateChange = kvp.Value.Value;
                        break;
                    case kTau:
                        tau = kvp.Value.Value;
                        break;
                }
            }
        }

        public Dictionary<string, RangedValue> GetParameters()
        {
            Dictionary<string, RangedValue> parameters = new Dictionary<string, RangedValue>();
            parameters.Add(kPopulationSize, new RangedValue(populationSize, 3, int.MaxValue, true));
            parameters.Add(kExitFitness, new RangedValue((float)exitFitness, 0, float.MaxValue, false));
            parameters.Add(kMaxGenerations, new RangedValue(maxGenerations, 0, int.MaxValue, true));
            parameters.Add(kMutateRate, new RangedValue((float)mutateRate, 0, 1, false));
            parameters.Add(kMutateChange, new RangedValue((float)mutateChange, 0, 1, false));
            parameters.Add(kTau, new RangedValue((float)tau, 0, 1, false));
            return parameters;
        }

        public RangedValue[] Solve(Func<RangedValue[], FitnessObject> fitnessFunction, RangedValue[] initialValues)
        {
            // set up parameters
            bool optimizationDone = false;
            int generation = 0;
            double bestFitness = double.MaxValue;
            RangedValue[] bestSolution = new RangedValue[initialValues.Length];

            // initialize a population of random solutions
            for (int i = 0; i < populationSize; i++)
            {
                Individual individual = new Individual(Optimizer.CreateRandomParameters(initialValues), 0);
                individual.fitness = fitnessFunction(individual.chromosome).fitness;
                population[i] = individual;
                if (individual.fitness < bestFitness)
                {
                    bestSolution = RangedValue.CopyArray(individual.chromosome);
                    bestFitness = individual.fitness;
                }
                Console.WriteLine("Initialize individual " + i + ": " + individual);
            }

            // loop until done
            while (!optimizationDone && (maxGenerations == 0 || generation < maxGenerations))
            {
                // select two good parent solutions
                Individual[] parents = select(2);

                // use them to generate two child solutions
                Individual[] children = crossover(parents[0], parents[1]);
                children[0] = mutate(children[0]);
                children[1] = mutate(children[1]);

                // place children into population
                Array.Sort(population);
                population[populationSize - 1] = children[0];
                population[populationSize - 2] = children[1];

                // immigrate a new solution into population
                Individual immigrant = new Individual(Optimizer.CreateRandomParameters(initialValues), 0);
                population[populationSize - 3] = immigrant;

                // Check new individuals
                for (int i = populationSize - 3; i < populationSize; i++)
                {
                    population[i].fitness = fitnessFunction(population[i].chromosome).fitness;
                    if (population[i].fitness < bestFitness)
                    {
                        bestFitness = population[i].fitness;
                        bestSolution = RangedValue.CopyArray(population[i].chromosome);
                        if (bestFitness < exitFitness)
                        {
                            optimizationDone = true;
                        }
                    }
                }
                Console.WriteLine("new individuals: " + population[populationSize - 1] + ", " +
                    population[populationSize - 2] + ", " + population[populationSize - 3]);
                Console.WriteLine("Generation: " + generation + ", best individual: " + new Individual(bestSolution, bestFitness));
                ++generation;
            }

            // return best solution found
            return bestSolution;
        }

        private Individual[] select(int n)
        {
            int[] indexes = new int[populationSize];
            for (int i = 0; i < indexes.Length; ++i)
                indexes[i] = i;

            for (int i = 0; i < indexes.Length; ++i) // shuffle
            {
                int r = rnd.Next(i, indexes.Length);
                int tmp = indexes[r]; indexes[r] = indexes[i]; indexes[i] = tmp;
            }

            int tournSize = (int)(tau * populationSize);
            if (tournSize < n) tournSize = n;
            Individual[] candidates = new Individual[tournSize];

            for (int i = 0; i < tournSize; ++i)
                candidates[i] = population[indexes[i]];
            Array.Sort(candidates);

            Individual[] results = new Individual[n];
            for (int i = 0; i < n; ++i)
                results[i] = candidates[i].Clone();

            return results;
        }

        private Individual mutate(Individual child)
        {
            for (int i = 0; i < child.chromosome.Length; ++i)
            {
                double hi = mutateChange * ((float)child.chromosome[i].Max);
                if (child.chromosome[i].IsInt && hi < 1)
                {
                    hi = 1;
                }
                double lo = -hi;
                if (rnd.NextDouble() < mutateRate)
                {
                    RangedValue gene = child.chromosome[i];
                    double delta = (hi - lo) * rnd.NextDouble() + lo;
                    double newValue = (float)gene.Value + delta;
                    if (newValue < gene.Min)
                    {
                        newValue = gene.Min;
                    }
                    else if (newValue > gene.Max)
                    {
                        newValue = gene.Max;
                    }
                    gene.Value = (float)newValue;
                }
            }
            return child;
        }

        private Individual[] crossover(Individual parent1, Individual parent2)
        {
            int numGenes = parent1.chromosome.Length;

            int cross = rnd.Next(0, numGenes - 1); // crossover point. 0 means 'between 0 and 1'.

            Individual child1 = new Individual(RangedValue.CopyArray(parent1.chromosome), 0);
            Individual child2 = new Individual(RangedValue.CopyArray(parent2.chromosome), 0);

            for (int i = 0; i <= cross; ++i)
                child1.chromosome[i] = parent1.chromosome[i].Clone();
            for (int i = cross + 1; i < numGenes; ++i)
                child2.chromosome[i] = parent1.chromosome[i].Clone();
            for (int i = 0; i <= cross; ++i)
                child2.chromosome[i] = parent2.chromosome[i].Clone();
            for (int i = cross + 1; i < numGenes; ++i)
                child1.chromosome[i] = parent2.chromosome[i].Clone();

            return new Individual[] { child1, child2 };
        }

        private class Individual : IComparable<Individual>
        {
            public RangedValue[] chromosome;
            public double fitness;

            public Individual(RangedValue[] values, double fitness)
            {
                chromosome = values;
                this.fitness = fitness;
            }

            public int CompareTo(Individual other)
            {
                if (this.fitness < other.fitness) return -1;
                else if (this.fitness > other.fitness) return 1;
                else return 0;
            }

            public Individual Clone()
            {
                RangedValue[] newChromosome = RangedValue.CopyArray(chromosome);
                return new Individual(newChromosome, fitness);
            }

            public override string ToString()
            {
                string result = "Individual[chromosome = (";
                for (int i = 0; i < chromosome.Length; i++)
                {
                    result += chromosome[i].Value;
                    if (i != chromosome.Length - 1)
                    {
                        result += ",";
                    }
                }
                result += "), fitness = " + fitness + "]";
                return result;
            }
        }

        public void SetParamNames(List<string> paramNames)
        {
            this.paramNames = paramNames;
        }

        public List<DeutscheBahn.OptimizationResult> getOptimizationResults()
        {
            return null;
        }
    }
}
