﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace Framework.optimization
{
    class NSGA2 : IOptimizationAlgorithm
    {
        const string kPopulationSize = "populationSize";
        const string kExitFitness = "exitFitness";
        const string kExitFitnessCount = "exitFitnessCount";
        const string kMaxGenerations = "maxGenerations";
        const string kMutateRate = "mutateRate";
        const string kMutateChange = "mutateChange";
        const string kTau = "tau";

        Random rnd = new Random();

        List<Individual> population;
        List<List<Individual>> fronts;
        List<DeutscheBahn.OptimizationResult> optimizationResults;

        /// <summary>
        /// Size of the population
        /// </summary>
        int populationSize = 500;
        /// <summary>
        /// Optimization is finished when fitness hasn't changed more than exitFitness the last exitFitnessCount times
        /// </summary>
        double exitFitness = 0.001;
        /// <summary>
        /// Counter for the exitFitness. Set to 0 to disable exit fitness.
        /// </summary>
        int exitFitnessCount = 10;
        /// <summary>
        /// Optimization will abort after this many loops
        /// </summary>
        int maxGenerations = 100;
        /// <summary>
        /// Likelihood that a gene is changed
        /// </summary> 
        double mutateRate = 0.60;
        /// <summary>
        /// Controls magnitude of mutation change as percentage of total gene range
        /// </summary>
        double mutateChange = 0.4;
        /// <summary>
        /// Selection pressure (percentage of population selected for tournament selection)
        /// </summary>
        double tau = 0.40;

        private List<string> paramNames;

        public NSGA2()
        {
            population = new List<Individual>(populationSize);
            fronts = new List<List<Individual>>();
            fronts.Add(new List<Individual>());
            optimizationResults = new List<DeutscheBahn.OptimizationResult>();
        }

        public void SetParameters(Dictionary<string, RangedValue> parameters)
        {
            foreach (KeyValuePair<string, RangedValue> kvp in parameters)
            {
                switch (kvp.Key)
                {
                    case kPopulationSize:
                        populationSize = (int)kvp.Value.Value;
                        population = new List<Individual>(populationSize);
                        break;
                    case kExitFitness:
                        exitFitness = kvp.Value.Value;
                        break;
                    case kExitFitnessCount:
                        exitFitnessCount = (int)kvp.Value.Value;
                        break;
                    case kMaxGenerations:
                        maxGenerations = (int)kvp.Value.Value;
                        break;
                    case kMutateRate:
                        mutateRate = kvp.Value.Value;
                        break;
                    case kMutateChange:
                        mutateChange = kvp.Value.Value;
                        break;
                    case kTau:
                        tau = kvp.Value.Value;
                        break;
                }
            }
        }

        public Dictionary<string, RangedValue> GetParameters()
        {
            Dictionary<string, RangedValue> parameters = new Dictionary<string, RangedValue>();
            parameters.Add(kPopulationSize, new RangedValue(populationSize, 3, int.MaxValue, true));
            parameters.Add(kExitFitness, new RangedValue((float)exitFitness, 0, float.MaxValue, false));
            parameters.Add(kExitFitnessCount, new RangedValue(exitFitnessCount, 0, int.MaxValue, true));
            parameters.Add(kMaxGenerations, new RangedValue(maxGenerations, 0, int.MaxValue, true));
            parameters.Add(kMutateRate, new RangedValue((float)mutateRate, 0, 1, false));
            parameters.Add(kMutateChange, new RangedValue((float)mutateChange, 0, 1, false));
            //parameters.Add(kTau, new RangedValue((float)tau, 0, 1, false));
            return parameters;
        }

        public RangedValue[] Solve(Func<RangedValue[], FitnessObject> fitnessFunction, RangedValue[] initialValues)
        {
            Console.WriteLine("Called NSGA2");
            Thread.CurrentThread.CurrentCulture = new System.Globalization.CultureInfo("en-Us");

            // initialize a population of random solutions
            for (int i = 0; i < populationSize; i++)
            {
                Individual individual = new Individual(Optimizer.CreateRandomParameters(initialValues), 0);
                FitnessObject fitnessResult = fitnessFunction(individual.chromosome);

                individual.fitness = fitnessResult.fitness;
                individual.errTot = fitnessResult.errTot;
                individual.compRate = fitnessResult.CR;
                population.Add(individual);
                //Console.WriteLine("Initialize individual " + i + ": " + individual);
            }

            Console.WriteLine("Initialized Individuals");

            List<Individual> parents = new List<Individual>(population);
            List<Individual> offspring = new List<Individual>();
            Queue<double> fitnessExitQueue = new Queue<double>();

            Stopwatch stopWatchGeneration = new Stopwatch();

            for (int t = 0; t < maxGenerations; t++)
            {
                stopWatchGeneration.Restart();

                population = new List<Individual>(parents);
                population.AddRange(offspring);

                nonDominatedSort();
                Console.WriteLine("Finished non dominated sort");

                parents = new List<Individual>();
                int i = 0;

                while (i < fronts.Count && parents.Count + fronts[i].Count < populationSize)
                {
                    parents.AddRange(fronts[i]);
                    i++;
                }

                if (parents.Count < populationSize)
                {
                    crowdingDistance(i);
                    fronts[i].Sort();
                    parents.AddRange(fronts[i].GetRange(0, populationSize - parents.Count));
                }


                offspring = makeNewPop(parents, initialValues, fitnessFunction);
                Console.WriteLine("Finished new pop");

                stopWatchGeneration.Stop();
                Console.WriteLine("Calculated Generation " + t + " of size " + population.Count + " in " + stopWatchGeneration.Elapsed);

                /*double avgFitness = fronts[0].Average(x => x.fitness);

                if (avgFitness > 0 && exitFitnessCount > 0)
                {
                    if (fitnessExitQueue.Count == 0 || Math.Abs(fitnessExitQueue.Last() - avgFitness) <= exitFitness)
                        fitnessExitQueue.Enqueue(avgFitness);
                    else
                        fitnessExitQueue.Clear();

                    if (fitnessExitQueue.Count == exitFitnessCount)
                        break;
                } */
            }

            Console.WriteLine("########################################");

            if (fitnessExitQueue.Count == exitFitnessCount && exitFitnessCount > 0)
            {
                Console.Write("Stopped because exit fitness reached: ");
                fitnessExitQueue.ToList().ForEach(i => Console.Write(i + ","));
                Console.WriteLine();
            }
            else Console.WriteLine("Stopped because max generations reached");

            foreach (List<Individual> front in fronts)
            {
                foreach (Individual individual in front)
                {
                    DeutscheBahn.OptimizationResult optiResult = new DeutscheBahn.OptimizationResult();
                    optiResult.compressionratio = individual.compRate;
                    optiResult.errortotal = individual.errTot;
                    optiResult.rang = individual.rank;
                    string parameters = "";

                    for (int i = 0; i < individual.chromosome.Length; i++)
                    {
                        parameters += paramNames[i] + "=" + individual.chromosome[i].Value + ";";
                    }
                    parameters += "fitness=" + individual.fitness;

                    optiResult.@params = parameters;

                    optimizationResults.Add(optiResult);
                }
            }

            fronts[0].Sort(new IndividualFitnessComparer());

            foreach (Individual individual in fronts[0])
            {
                Console.WriteLine(individual);
            }


            Console.WriteLine("Best Individual: " + fronts[0][0]);

            //FitnessObject testResult = fitnessFunction(fronts[0][0].chromosome);
            //Console.WriteLine("Recalcuated fitness: " + testResult.fitness);

            return fronts[0][0].chromosome;
        }

        private List<Individual> makeNewPop(List<Individual> parent, RangedValue[] initialValues, Func<RangedValue[], FitnessObject> fitnessFunction)
        {
            parent.Sort();

            List<Individual> offspring = new List<Individual>();

            for (int i = 1; i < parent.Count; i += 2)
            {
                Individual[] children = crossover(parent[i - 1], parent[i]);
                offspring.Add(mutate(children[0]));
                offspring.Add(mutate(children[1]));
            }

            foreach (Individual individual in offspring)
            {
                Console.Write("\t");
                FitnessObject fitnessResult = fitnessFunction(individual.chromosome);

                individual.fitness = fitnessResult.fitness;
                individual.errTot = fitnessResult.errTot;
                individual.compRate = fitnessResult.CR;
            }

            return offspring;
        }

        private Individual mutate(Individual child)
        {
            for (int i = 0; i < child.chromosome.Length; ++i)
            {
                double hi = mutateChange * ((float)child.chromosome[i].Max);
                if (child.chromosome[i].IsInt && hi < 1)
                {
                    hi = 1;
                }
                double lo = -hi;

                RangedValue gene = child.chromosome[i];
                double delta = (hi - lo) * rnd.NextDouble() + lo;
                double newValue = (float)gene.Value + delta;
                if (newValue < gene.Min)
                {
                    newValue = gene.Min;
                }
                else if (newValue > gene.Max)
                {
                    newValue = gene.Max;
                }
                gene.Value = (float)newValue;

            }
            return child;
        }

        private Individual[] crossover(Individual parent1, Individual parent2)
        {
            int numGenes = parent1.chromosome.Length;

            int cross = rnd.Next(0, numGenes - 1); // crossover point. 0 means 'between 0 and 1'.

            Individual child1 = new Individual(RangedValue.CopyArray(parent1.chromosome), 0);
            Individual child2 = new Individual(RangedValue.CopyArray(parent2.chromosome), 0);

            for (int i = 0; i <= cross; ++i)
                child1.chromosome[i] = parent1.chromosome[i].Clone();
            for (int i = cross + 1; i < numGenes; ++i)
                child2.chromosome[i] = parent1.chromosome[i].Clone();
            for (int i = 0; i <= cross; ++i)
                child2.chromosome[i] = parent2.chromosome[i].Clone();
            for (int i = cross + 1; i < numGenes; ++i)
                child1.chromosome[i] = parent2.chromosome[i].Clone();

            return new Individual[] { child1, child2 };
        }

        private void nonDominatedSort()
        {
            fronts = new List<List<Individual>>();
            fronts.Add(new List<Individual>());

            foreach (Individual p in population)
            {
                p.dominatingIndividuals = new List<Individual>();
                p.dominationCounter = 0;

                // search for individuals which are dominating p or which are dominated by p
                foreach (Individual q in population)
                {
                    if (p.isDominating(q))
                        p.dominatingIndividuals.Add(q);
                    else if (q.isDominating(p))
                        p.dominationCounter++;
                }

                // if there's no individual dominating p, add p to the first front
                if (p.dominationCounter == 0)
                {
                    p.rank = 0;

                    //if (!fronts[0].Contains(p))
                    fronts[0].Add(p);
                }
            }

            //Console.WriteLine("Front Count: " + fronts.Count);

            for (int i = 0; i < fronts.Count; i++)
            {
                List<Individual> newFront = new List<Individual>();

                //Console.WriteLine("Front " + i + " Count: " + fronts[i].Count);

                foreach (Individual p in fronts[i])
                {
                    foreach (Individual q in p.dominatingIndividuals)
                    {
                        q.dominationCounter--;

                        if (q.dominationCounter == 0)
                        {
                            q.rank = i + 1;
                            newFront.Add(q);

                            //Console.WriteLine("Add to new Front " + (i + 1));
                        }
                    }
                }

                //Console.WriteLine("Add new Front: " + (fronts.Count <= i + 1));

                if (newFront.Count > 0 && fronts.Count <= i + 1)
                    fronts.Add(newFront);
                else if (fronts.Count < i + 1)
                    fronts[i + 1] = newFront;
            }
        }

        private void crowdingDistance(int iFront)
        {
            List<Individual> front = fronts[iFront];
            List<Individual> errTotList = new List<Individual>(front);
            List<Individual> crList = new List<Individual>(front);

            errTotList.Sort(new IndividualErrTotComparer());
            crList.Sort(new IndividualCompRateComparer());

            for (int i = 0; i < front.Count; i++)
            {
                if (i == 0 || i == front.Count - 1)
                {
                    errTotList[i].crowdingDistance = Double.PositiveInfinity;
                    crList[i].crowdingDistance = Double.PositiveInfinity;
                }
                else
                {
                    errTotList[i].crowdingDistance += (errTotList[i + 1].errTot - errTotList[i - 1].errTot) / 5;
                    crList[i].crowdingDistance += (crList[i + 1].compRate - crList[i - 1].compRate) / 2;
                }
            }
        }

        private class Individual : IComparable<Individual>
        {
            public RangedValue[] chromosome;
            public double fitness;
            public double errTot;
            public double compRate;

            public int rank;
            public double crowdingDistance;
            public List<Individual> dominatingIndividuals;
            public int dominationCounter;

            public Individual(RangedValue[] chromosome, double fitness)
                : this(chromosome, fitness, -1, -1)
            {
            }

            public Individual(RangedValue[] chromosome, double fitness, int rank, double crowdingDistance)
            {
                this.chromosome = chromosome;
                this.fitness = fitness;
                this.rank = rank;
                this.crowdingDistance = crowdingDistance;
                this.dominatingIndividuals = new List<Individual>();
                this.dominationCounter = 0;
            }

            public int CompareTo(Individual other)
            {
                if (this.rank < other.rank) return -1;
                else if (this.rank > other.rank) return 1;
                else
                    if (this.crowdingDistance < other.crowdingDistance) return -1;
                    else if (this.crowdingDistance > other.crowdingDistance) return 1;
                    else return 0;
            }

            public override string ToString()
            {
                string result = "Individual[rank = " + rank + ", crowding distance = " + crowdingDistance + ", errTot = " + errTot + ", compRate = " + compRate + ", fitness = " + fitness + " chromosome = (";

                for (int i = 0; i < chromosome.Length; i++)
                {
                    result += chromosome[i].Value;
                    if (i != chromosome.Length - 1)
                    {
                        result += ",";
                    }
                }
                result += ")]";

                return result;
            }

            public Boolean isDominating(Individual other)
            {
                if (this.errTot < other.errTot && this.compRate < other.compRate)
                    return true;
                return false;
            }
        }

        private class IndividualFitnessComparer : IComparer<Individual>
        {
            public int Compare(Individual a, Individual b)
            {
                return a.fitness.CompareTo(b.fitness);
            }
        }

        private class IndividualErrTotComparer : IComparer<Individual>
        {
            public int Compare(Individual a, Individual b)
            {
                return a.errTot.CompareTo(b.errTot);
            }
        }

        private class IndividualCompRateComparer : IComparer<Individual>
        {
            public int Compare(Individual a, Individual b)
            {
                return a.compRate.CompareTo(b.compRate);
            }
        }

        public void SetParamNames(List<string> paramNames)
        {
            this.paramNames = paramNames;
        }

        public List<DeutscheBahn.OptimizationResult> getOptimizationResults()
        {
            return optimizationResults;
        }
    }
}
