﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Framework.optimization
{
    class NelderMeadSolver : IOptimizationAlgorithm
    {
        const string kMaxIterations = "maxIterations";
        const string kExitFitness = "exitFitnes";
        const string kAlpha = "alpha";
        const string kBeta = "beta";
        const string kGamma = "gamma";

        int maxIterations = 10000;
        float exitFitness = 0.5f;
        float alpha = 1, beta = 0.5f, gamma = 2f;

        public void SetParameters(Dictionary<string, RangedValue> parameters)
        {
            if (parameters.ContainsKey(kAlpha)) alpha = parameters[kAlpha].Value;
            if (parameters.ContainsKey(kBeta)) beta = parameters[kBeta].Value;
            if (parameters.ContainsKey(kGamma)) gamma = parameters[kGamma].Value;
            if (parameters.ContainsKey(kMaxIterations)) maxIterations = (int)parameters[kMaxIterations].Value;
            if (parameters.ContainsKey(kExitFitness)) exitFitness = parameters[kExitFitness].Value;
        }

        public Dictionary<string, RangedValue> GetParameters()
        {
            Dictionary<string, RangedValue> parameters = new Dictionary<string, RangedValue>();
            parameters.Add(kAlpha, new RangedFloatValue(alpha, 0, float.MaxValue));
            parameters.Add(kBeta, new RangedFloatValue(beta, 0, float.MaxValue));
            parameters.Add(kGamma, new RangedFloatValue(gamma, 0, float.MaxValue));
            parameters.Add(kMaxIterations, new RangedIntValue(maxIterations, 0, int.MaxValue));
            parameters.Add(kExitFitness, new RangedFloatValue(exitFitness, 0, float.MaxValue));
            return parameters;
        }

        private RangedValue[] Reflect(RangedValue[] xmax, RangedValue[] xmean, float alpha)
        {
            RangedValue[] result = RangedValue.CopyArray(xmean);
            for (int i = 0; i < xmean.Length; i++)
            {
                result[i].Value = EvaluateValue((1 + alpha) * (xmean[i].Value) - alpha * (xmax[i].Value), result[i]);
            }
            return result;
        }

        private RangedValue[] Expanse(RangedValue[] xr, RangedValue[] xmean, float gamma)
        {
            RangedValue[] result = RangedValue.CopyArray(xmean);
            for (int i = 0; i < xr.Length; i++)
            {
                result[i].Value = EvaluateValue((1 + gamma) * (xr[i].Value) - gamma * (xmean[i].Value), result[i]);
            }
            return result;
        }

        private float EvaluateValue(float val, RangedValue rv)
        {
            if (val > rv.Max) return rv.Max;
            else if (val < rv.Min) return rv.Min;
            else return val;
        }

        private RangedValue[] Contract(RangedValue[] xmax, RangedValue[] xmean, float beta)
        {
            RangedValue[] result = RangedValue.CopyArray(xmax);

            for (int i = 0; i < xmax.Length; i++)
            {
                result[i].Value = EvaluateValue((beta) * (xmax[i].Value) + (1 - beta) * (xmean[i].Value), result[i]);
            }
            return result;
        }

        private RangedValue[][] CompressSimplex(RangedValue[][] xi, RangedValue[] xmin)
        {
            Console.WriteLine("===== Compress Simplex =====");
            RangedValue[][] result = new RangedValue[xi.Length][];
            for (int i = 0; i < xi.Length; i++)
            {
                result[i] = RangedValue.CopyArray(xmin);
                for (int j = 0; j < result[i].Length; j++)
                {
                    result[i][j].Value = EvaluateValue((xi[i][j].Value + xmin[j].Value) / 2, result[i][j]);
                }
            }
            return result;
        }

        private RangedValue[] GetMean(RangedValue[][] xi)
        {
            float[] mean = new float[xi[0].Length];
            for (int i = 0; i < xi[0].Length; i++)
            {
                //Init
                mean[i] = 0;
            }
            int N = xi.Length;
            //Sum up Mean
            foreach (RangedValue[] x in xi)
            {
                for (int i = 0; i < x.Length; i++)
                    mean[i] += x[i].Value / N;
            }
            //write back to parameters
            RangedValue[] xmean = RangedValue.CopyArray(xi[0]);
            for (int i = 0; i < xmean.Length; i++)
            {
                xmean[i].Value = EvaluateValue(mean[i], xmean[i]);
            }
            return xmean;
        }


        RangedValue[][] InitParams(RangedValue[] init)
        {
            RangedValue[][] XInit = new RangedValue[init.Length + 1][];
            for (int i = 0; i < init.Length + 1; i++)
            {
                XInit[i] = Optimizer.CreateRandomParameters(init);
            }
            return XInit;
        }

        RangedValue[] IOptimizationAlgorithm.Solve(Func<RangedValue[], FitnessObject> fitnessFunction, RangedValue[] initialValues)
        {
            Console.WriteLine(alpha + " " + beta + " " + gamma);
            int N = initialValues.Length;
            //Erzeuge N+1 Lösungen, jede Lösung enthält soviele Variablen wie InitialValues Werte hat
            RangedValue[][] XI = InitParams(initialValues);//initialValues.Length

            //This queue save the last 10 FitnessChanges
            FixedSizedQueue<double> lastFitnessChanges = new FixedSizedQueue<double>(10);
            double lastymin = double.MaxValue;

            int xmaxi = 0;
            int xmini = 0;
            double ymax = double.MinValue;
            double ymin = double.MaxValue;
            int iterations = 0;
            Random r = new Random();

            while (!(ymin < exitFitness || iterations > maxIterations))
            {
                Console.WriteLine("Iteration : " + iterations++);

                //Yi für alle Xi setzen
                double[] yi = new double[XI.Length];
                for (int i = 0; i < XI.Length; i++)
                {
                    yi[i] = fitnessFunction(XI[i]).fitness;
                }

                ymax = double.MinValue;
                ymin = double.MaxValue;
                //Bestimme alle Maxima
                for (int i = 0; i < yi.Length; i++)
                {
                    if (yi[i] < ymin) { ymin = yi[i]; xmini = i; }
                    if (yi[i] > ymax) { ymax = yi[i]; xmaxi = i; }
                }

                //Save the last ymin
                lastFitnessChanges.Enqueue(lastymin - ymin);
                lastymin = ymin;

                if (lastFitnessChanges.Count == lastFitnessChanges.Size && lastFitnessChanges.Sum() <= 0)
                {
                    Console.WriteLine("======= Fitness didnt get better ... ymin = " + ymin);
                    int parami = 0;
                    for (int i = 0; i < XI.Length; i++)
                    {
                        if (i != xmini)
                        {
                            //Create new Random values for all Values but the best
                            //XI[i] = Optimizer.CreateRandomParameters(initialValues);

                            // Create a variation of the minimum
                            XI[i] = RangedValue.CopyArray(XI[xmini]);
                            XI[i][parami].Value = EvaluateValue(XI[i][parami].Value + XI[i][parami].Value * 0.3f * (r.Next(2) == 0 ? 1 : -1), XI[i][parami]);
                            parami++;
                        }
                    }
                    lastFitnessChanges = new FixedSizedQueue<double>(10);
                    continue; // to calculate fitness of these new values
                }

                //Berechne Mittelwert X-Quer
                RangedValue[] xmean = GetMean(XI);

                //Berechne Reflexion
                RangedValue[] xr = Reflect(XI[xmaxi], xmean, alpha);
                Console.WriteLine("Reflect: ");
                double yr = fitnessFunction(xr).fitness;

                if (yr < ymin)
                {
                    //Berechne Expansion
                    RangedValue[] xe = Expanse(xr, xmean, gamma);
                    Console.WriteLine("yr < ymin __ Expanse: ");
                    double ye = fitnessFunction(xe).fitness;
                    if (ye < ymin)
                    {
                        Console.WriteLine("ye < ymin __ accept expansion");
                        //Akzeptiere Expansion
                        XI[xmaxi] = RangedValue.CopyArray(xe);
                    }
                    else
                    {
                        Console.WriteLine("ye >= ymin __ accept reflection");
                        //Akzeptiere Reflexion
                        XI[xmaxi] = RangedValue.CopyArray(xr);
                    }
                }
                else
                {
                    Console.WriteLine("yr >= ymin __ look for xi != xmax, yr < yi");
                    //Gibt es xi!=xmax so dass yr < yi
                    bool yrsmalleryi = false;
                    for (int i = 0; i < XI.Length; i++)
                    {
                        if (yr < yi[i] && i != xmaxi)
                        {
                            yrsmalleryi = true;
                            break;
                        }
                    }
                    //Wenn es so eines gibt
                    if (yrsmalleryi)
                    {
                        Console.WriteLine("yr < yi __ accept reflection " + xmaxi);
                        //Akzeptiere Reflexion
                        XI[xmaxi] = RangedValue.CopyArray(xr);
                    }
                    else
                    {
                        Console.WriteLine("yr >= yi");
                        //yr < ymax
                        if (yr < ymax)
                        {
                            Console.WriteLine("yr < ymax __ replace xmax with xr");
                            //ersetze xmax durch xr
                            XI[xmaxi] = RangedValue.CopyArray(xr);
                            ymax = yr; //TODO: ist das notwendig, wird explizit nicht geschrieben (ganz oben rechts auf Flowchart)
                        }
                        //Berechne Kontraktion
                        RangedValue[] xcon = Contract(XI[xmaxi], xmean, beta);
                        Console.WriteLine("Contract: ");
                        double ycon = fitnessFunction(xcon).fitness;
                        if (ycon < ymax)
                        {
                            Console.WriteLine("ycon < ymax __ accept contraction");
                            //Akzeptiere Kontraktion
                            XI[xmaxi] = RangedValue.CopyArray(xcon);
                        }
                        else
                        {
                            Console.WriteLine("ycon >= ymax __ compress simplex");
                            //Komprimiere Simplex
                            XI = CompressSimplex(XI, XI[xmini]);
                        }
                    }
                }
            }

            return XI[xmini];
        }

        public class FixedSizedQueue<T> : ConcurrentQueue<T>
        {
            public int Size { get; private set; }

            public FixedSizedQueue(int size)
            {
                Size = size;
            }

            public new void Enqueue(T obj)
            {
                base.Enqueue(obj);
                lock (this)
                {
                    while (base.Count > Size)
                    {
                        T outObj;
                        base.TryDequeue(out outObj);
                    }
                }
            }
        }

        public void SetParamNames(List<string> paramNames) 
        {
 
        }

        public List<DeutscheBahn.OptimizationResult> getOptimizationResults()
        {
            return null;
        }
    }
}
