﻿using Framework.changedetection;
using Framework.compression;
using System;
using System.Collections.Generic;
using System.Linq;

namespace Framework.optimization
{
    class Optimizer
    {
        static Random rand = new Random();

        //Total Error aus Maß, vorfaktoren
        private double a = 1, b = 20, c = 4;

        //VorFaktoren
        private double alpha = 1;
        private double beta = 5;

        private IChangeDetectionAlgorithm changedetection;
        private ICompressionAlgorithm compression;
        private IOptimizationAlgorithm optimization;

        //Saved data for comparison
        private float[] originaldata;
        private float[] originalscores;
        private int[] originalChangePoints;
        private Func<RangedValue[], FitnessObject> optimizeFunc;

        /// <summary>
        /// Creat Optimizer Object with ChangeDetection, Compression and Optimization Algorithm
        /// The Initial Parameters for the optimization Process are the setParameters of ChangeDetection and Compression
        /// ChangeDetection MUST have been executed before, the detected Changes are our golden standard for the optimization goal
        /// </summary>
        /// <param name="rawdata">Dataset</param>
        /// <param name="cdobject">Executed and parametrized ChangeDetection Algorithm</param>
        /// <param name="cmpobject">Parametrized Compression Algorithm</param>
        /// <param name="optiobj">Optimizaton Object </param>
        public Optimizer(float[] rawdata, IChangeDetectionAlgorithm cdobject, ICompressionAlgorithm cmpobject, IOptimizationAlgorithm optiobj)
            : this(rawdata, cdobject.GetScores(), cdobject.GetChangePoints(), cdobject, cmpobject, optiobj)
        {
        }
        public Optimizer(float[] rawdata, IChangeDetectionAlgorithm cdobject, IOptimizationAlgorithm optiobj)
            : this(rawdata, cdobject.GetScores(), cdobject.GetChangePoints(), cdobject, null, optiobj)
        {
        }

        public Optimizer(float[] rawdata, float[] originalscores, int[] originalChangePoints, IChangeDetectionAlgorithm cdobject, IOptimizationAlgorithm optiobj)
            : this(rawdata, originalscores, originalChangePoints, cdobject, null, optiobj)
        {
        }

        public Optimizer(float[] rawdata, float[] originalscores, int[] originalChangePoints, IChangeDetectionAlgorithm cdobject, ICompressionAlgorithm cmpobject, IOptimizationAlgorithm optiobj)
        {
            optimizeFunc = this.calculateFitness;

            this.changedetection = cdobject;
            this.compression = cmpobject;
            this.optimization = optiobj;
            originaldata = rawdata;

            if (null != originalscores)
            {
                this.originalscores = new float[originalscores.Length];
                originalscores.CopyTo(this.originalscores, 0);
            }

            this.originalChangePoints = new int[originalChangePoints.Length];
            originalChangePoints.CopyTo(this.originalChangePoints, 0);

        }

        public void SetOptimizationParameters(double alpha, double beta)
        {
            this.alpha = alpha;
            this.beta = beta;
        }

        public void SetMeasurementParameters(double a, double b, double c)
        {
            this.a = a;
            this.b = b;
            this.c = c;
        }

        public Dictionary<String, RangedValue> GetParameters()
        {
            Dictionary<String, RangedValue> parameters = new Dictionary<string, RangedValue>();
            parameters.Add("alpha", new RangedFloatValue((float)alpha, 0, float.MaxValue));
            parameters.Add("beta", new RangedFloatValue((float)beta, 0, float.MaxValue));
            parameters.Add("a", new RangedFloatValue((float)a, 0, float.MaxValue));
            parameters.Add("b", new RangedFloatValue((float)b, 0, float.MaxValue));
            parameters.Add("c", new RangedFloatValue((float)c, 0, float.MaxValue));
            return parameters;
        }

        public void SetParameters(Dictionary<string, RangedValue> parameters)
        {
            foreach (KeyValuePair<string, RangedValue> kvp in parameters)
            {
                switch (kvp.Key)
                {
                    case "alpha":
                        alpha = kvp.Value.Value;
                        break;
                    case "beta":
                        beta = kvp.Value.Value;
                        break;
                    case "a":
                        a = kvp.Value.Value;
                        break;
                    case "b":
                        b = kvp.Value.Value;
                        break;
                    case "c":
                        c = kvp.Value.Value;
                        break;
                }
            }
        }

        public RangedValue[] Solve()
        {
            List<RangedValue> initChangeDetection = new List<RangedValue>();
            List<string> paramNames = new List<string>();
            if (changedetection != null)
            {
                // HACK to correctly initialize some RangedValue limits. Assumes that original and decompressed data have same length.
                changedetection.SetDataset(originaldata);
                foreach (KeyValuePair<string, RangedValue> kvp in changedetection.GetParameters())
                {
                    initChangeDetection.Add(kvp.Value);
                    paramNames.Add(kvp.Key);
                }
            }
            if (compression != null)
            {
                // HACK to correctly initialize some RangedValue limits. Assumes that original and decompressed data have same length.
                compression.SetDataset(originaldata);
                foreach (KeyValuePair<string, RangedValue> kvp in compression.GetParameters())
                {
                    initChangeDetection.Add(kvp.Value);
                    paramNames.Add(kvp.Key);
                }
            }

            //SOLVE
            optimization.SetParamNames(paramNames);
            RangedValue[] solution = optimization.Solve(optimizeFunc, initChangeDetection.ToArray());

            return solution;
        }

        public DeutscheBahn.JobResult createResult(RangedValue[] solution)
        {

            DeutscheBahn.JobResult result = new DeutscheBahn.JobResult();
            //Create Parameters
            Dictionary<String, RangedValue> myps = new Dictionary<String, RangedValue>();
            String pms = "";
            for (int i = 0; i < changedetection.GetParameters().Count; i++)
            {
                String name = changedetection.GetParameters().ToArray()[i].Key;
                pms += name + "=" + solution[i].Value + ";";
                myps.Add(name, solution[i]);
            }
            changedetection.SetParameters(myps);
            //SET RESULT =============================== CHANGE DETECTION PARAMS
            result.ChangeDetectionParameters = pms;

            float[] decompdata = originaldata;
            String pmss = "";
            if (compression != null)
            {
                //Set Parameters for Compression!
                Dictionary<String, RangedValue> compParams = new Dictionary<String, RangedValue>();
                //Set Values for Compression

                for (int i = 0; i < compression.GetParameters().Count; i++)
                {
                    String name = compression.GetParameters().ToArray()[i].Key;
                    pmss += name + "=" + solution[i + changedetection.GetParameters().Count].Value + ";";
                    compParams.Add(name, solution[i + changedetection.GetParameters().Count]);
                }
                compression.SetDataset(originaldata);
                compression.SetParameters(compParams);
                //Compress
                compression.CompressDataset();

                decompdata = compression.GetDecompressedDataset();
            }
            //SET RESULT =============================== COMPRESSION PARAMS
            result.CompressionParameters = pmss;

            //SET RESULT =============================== OPTIMIZATION RESULTS
            result.OptimizationResult = optimization.getOptimizationResults();

            //SET RESULT =============================== DECOMPRESSSION
            DeutscheBahn.DecompressedSet decomp = new DeutscheBahn.DecompressedSet();
            decomp.name = "yolo brolo";
            //decomp.decompressed_data = new List<DeutscheBahn.DecompressedData>();
            for (int i = 0; i < decompdata.Count(); i++)
            {
                DeutscheBahn.DecompressedData d = new DeutscheBahn.DecompressedData();
                d.time = i;
                d.value = decompdata[i];
                decomp.decompressed_data.Add(d);
            }
            result.decompressed_set = decomp;

            //Detect Changes on compressed Dataset which is decompressed ;)
            changedetection.SetDataset(decompdata);
            changedetection.DetectChanges();

            float[] compressedscores = changedetection.GetScores();
            int[] decompChangePoints = changedetection.GetChangePoints();

            //SET RESULT ================================ CHANGEPOINTS
            DeutscheBahn.ChangePointSet cps = new DeutscheBahn.ChangePointSet();
            //cps.change_point = new List<DeutscheBahn.ChangePoint>();
            foreach (int cp in changedetection.GetChangePoints())
            {
                DeutscheBahn.ChangePoint change = new DeutscheBahn.ChangePoint();
                change.time = cp;
                change.value = changedetection.GetScores()[cp];
                cps.change_point.Add(change);
            }
            result.changepoint_set = cps;


            //Compare
            ChangeDetectionMassLib.MeasureResult measure = ChangeDetectionMassLib.ChangeDetectionMass.get().calculateMeasure(originaldata, originalChangePoints, null, decompdata, decompChangePoints, null);

            //ScoreFitness!
            result.errFP = (float)measure.getFpError();
            result.errMISS = (float)measure.getMissError();
            result.errPC = (float)measure.getPcError();
            result.FP = measure.getNumFP();
            result.MISS = measure.getNumMISS();
            result.PC = measure.getNumPC();

            result.RMSE = Utilities.calculateRMSE(originaldata, decompdata);

            double errTot = getTotalError(measure);
            double CR = 0;
            if (compression != null)
                CR = compression.getCompressionRatio();

            result.CompressionRatio = (float)CR;
            result.Fitness = (float)getFitness(errTot, CR);

            return result;
        }

        public static RangedValue[] CreateRandomParameters(RangedValue[] parameters)
        {
            RangedValue[] randomParameters = RangedValue.CopyArray(parameters);
            for (int i = 0; i < randomParameters.Length; i++)
            {
                double range = randomParameters[i].Max - randomParameters[i].Min;
                randomParameters[i].Value = (float)(randomParameters[i].Min + rand.NextDouble() * range);
            }
            return randomParameters;
        }

        private FitnessObject calculateFitness(RangedValue[] x)
        {
            //Output for 
            String output = "";
            if (changedetection != null)
            {
                //Set Parameters for CD Rerun
                //Now iterate through given params and map them to Parameters of Algorithm
                Dictionary<String, RangedValue> myps = new Dictionary<String, RangedValue>();
                for (int i = 0; i < changedetection.GetParameters().Count; i++)
                {
                    String name = changedetection.GetParameters().ToArray()[i].Key;
                    //Output
                    output += (name + "=" + x[i].Value + " ");
                    Console.Write(name + "=" + x[i].Value + " ");
                    myps.Add(name, x[i]);
                }
                changedetection.SetParameters(myps);
            }
            //If no compression is used, original data is used. ... case we optimize 
            float[] decompdata = new float[originaldata.Length];
            originaldata.CopyTo(decompdata, 0);
            output += " | ";
            if (compression != null)
            {
                //Set Parameters for Compression!
                Dictionary<String, RangedValue> compParams = new Dictionary<String, RangedValue>();
                //Set Values for Compression
                for (int i = 0; i < compression.GetParameters().Count; i++)
                {
                    String name = compression.GetParameters().ToArray()[i].Key;
                    //Output
                    int offset = 0;
                    if (changedetection != null)
                        offset = changedetection.GetParameters().Count;
                    Console.Write(name + "=" + x[i + offset].Value + " ");
                    output += (name + "=" + x[i + offset].Value + " ");
                    compParams.Add(name, x[i + offset]);
                }
                Program.lastFitnessParams = output;
                compression.SetDataset(originaldata);
                compression.SetParameters(compParams);
                //Compress Data
                compression.CompressDataset();
                decompdata = compression.GetDecompressedDataset();
            }

            //Detect Changes on compressed Dataset which is decompressed ;)
            changedetection.SetDataset(decompdata);
            changedetection.DetectChanges();

            float[] compressedscores = changedetection.GetScores();
            int[] decompChangePoints = changedetection.GetChangePoints();

            //Normalize Compressed Scores if they have values >1 and originalScores hasn't (groundtruth vs cp algorithm)
            float compressedMax = compressedscores.Max();
            if (originalscores.Max() <= 1 && compressedMax > 1)
            {
                compressedscores = compressedscores.Select((element) => element / compressedMax).ToArray();
            }

            //Compare
            ChangeDetectionMassLib.MeasureResult measure = ChangeDetectionMassLib.ChangeDetectionMass.get().calculateMeasure(originaldata, originalChangePoints, null, decompdata, decompChangePoints, null);

            //ScoreFitness!
            //Console.WriteLine("|PC|=" + measure.getNumPC() + " |MISS|=" + measure.getNumMISS() + " |FP|=" + measure.getNumFP());
            double errTot = getTotalError(measure);
            double CR = 0;
            if (compression != null)
                CR = compression.getCompressionRatio();

            double totFitness = getFitness(errTot, CR);
            //Console.WriteLine("Z=" + Z + ", fz(Z)=" + fz + ", errTot=" + errTot + ", CR=" + CR);
            Console.WriteLine("Fitness : " + totFitness);
            return new FitnessObject((float)errTot, (float)CR, (float)getFitness(errTot, CR));
        }

        //Total error Computation as proposed by Engelhardt...
        public double getTotalError(ChangeDetectionMassLib.MeasureResult mr)
        {
            return (a * mr.getPcError() + b * mr.getMissError() + c * mr.getFpError()) / (mr.getNumPC() + mr.getNumMISS() + 1.0);
        }

        //Returns the fitness calculated as weighted Sum of errTot, CR and fZ
        public double getFitness(double errTot, double CR)
        {
            return alpha * errTot + beta * CR;
        }
    }
}
